import logging
from gremux.cmds import sessionizer
import gremux.cmds.places as plc
import gremux.cmds.config as cfg
import gremux.struct as struct


def main():
    import argparse

    parser = argparse.ArgumentParser(
        prog="gremux",
        description=(
            "Declarative tmux session manager.\n\n"
            "If run without any arguments, gremux launches the interactive "
            "sessionizer to select a project and attach to a tmux session."
        ),
    )

    sub = parser.add_subparsers(
        dest="cmd",
        metavar="<command>",
    )

    # ================================
    # Shared parent arguments
    # ================================
    source_parent = argparse.ArgumentParser(add_help=False)
    source_parent.add_argument(
        "--source",
        "-s",
        metavar="PATH",
        help=("Source file or directory.\nDefaults to the current working directory."),
    )

    # ================================
    # CONFIG COMMANDS
    # ================================
    config = sub.add_parser(
        "config",
        help="Manage gremux session configuration files",
        description=(
            "Commands for creating, inspecting, and launching tmux sessions "
            "from grem.yaml configuration files."
        ),
    )
    config_sub = config.add_subparsers(
        dest="config_cmd",
        metavar="<subcommand>",
        required=True,
    )

    config_sub.add_parser(
        "up",
        parents=[source_parent],
        help="Launch a tmux session from a grem.yaml file",
        description=(
            "Parse a grem.yaml configuration file and launch or attach to "
            "the corresponding tmux session."
        ),
    )

    config_sub.add_parser(
        "show",
        parents=[source_parent],
        help="Show the resolved configuration that will be used",
        description=(
            "Print the effective configuration after resolving defaults and paths."
        ),
    )

    config_sub.add_parser(
        "create",
        parents=[source_parent],
        help="Create a default grem.yaml configuration",
        description=(
            "Create a default grem.yaml configuration file.\n\n"
            "The file is created under ~/.config/gremux/default.yaml."
        ),
    )

    # ================================
    # PLACES COMMANDS
    # ================================
    places = sub.add_parser(
        "places",
        help="Manage common project locations",
        description=(
            "Commands for managing places.yaml, which defines common "
            "directories used by the sessionizer."
        ),
    )
    places_sub = places.add_subparsers(
        dest="places_cmd",
        metavar="<subcommand>",
        required=True,
    )

    places_create = places_sub.add_parser(
        "create",
        help="Create or update the places.yaml file",
        description=(
            "Create the places.yaml configuration file, optionally populating "
            "it from an external source such as zoxide."
        ),
    )

    places_create.add_argument(
        "--source",
        "-s",
        metavar="STR",
        help="default, backup, or zoxide",
    )

    places_create.add_argument(
        "--add",
        "-a",
        nargs="+",
        metavar="DIR",
        help="Add one or more directories to places.yaml",
    )

    places_create.add_argument(
        "--maximum",
        "-m",
        type=int,
        metavar="N",
        help="Maximum number of entries to include when using a dynamic source",
    )

    args = parser.parse_args()

    logger = struct.get_logger(level=logging.INFO)

    if args.cmd == "config":
        if args.config_cmd == "up":
            cfg.up(args, logger)
        elif args.config_cmd == "show":
            cfg.show(args, logger)
        elif args.config_cmd == "create":
            cfg.create(args, logger)

    elif args.cmd == "places":
        if args.places_cmd == "create":
            plc.create(args, logger)

    else:
        # Default behavior: sessionizer
        sessionizer(logger)



import yaml
from pathlib import Path
from gremux.struct.grem import Grem
from gremux.struct.context import Window, Pane, Layout


class Parser:
    """
    Base parser class
    """

    def __init__(self, path: Path | str = Path.cwd(), filename="grem.yaml"):
        self.path = path
        self.name = filename
        self.gremfile = Path(self.path).expanduser().resolve() / Path(self.name)

        self.yaml = None
        self.session_name = None

    def load(self) -> dict:
        """
        Load a grem.yaml file
        """

        if not self.gremfile.exists():
            raise NotADirectoryError("grem.yaml not found. Exiting...")

        with self.gremfile.open("r") as f:
            self.yaml = yaml.safe_load(f)

    def _default_grem(self) -> Grem:
        """
        Set the default to open in case there is no grem.yaml in the project's root.
        """
        panes = [Pane(command=None, cwd=None)]
        windows = [Window(name="default", layout=Layout(None), panes=panes)]
        grem = Grem(name=Path(self.path).name, windows=windows)

        return grem

    def grem(self) -> Grem:
        """
        From a given grem.yaml file, convert it into a configuration Grem object
        """

        # If there is no local grem.yaml, goes to default.yaml
        # If there is none, then it goes to a default thing still.

        if not self.gremfile.exists():
            config_path = Path.home() / ".config" / "gremux"
            filename = Path("default.yaml")
            default_gremfile = config_path / filename

            # If we're already trying the default, stop recursing
            if self.gremfile.resolve() == default_gremfile.resolve():
                return self._default_grem()

            # Otherwise, switch to default and try again
            self.session_name = Path(self.path).name
            self.name = filename
            self.gremfile = default_gremfile.expanduser().resolve()

            return self.grem()  # IMPORTANT: return it

        if self.yaml is None:
            self.load()

        session = self.yaml["session"]
        # Note: i'll hardcode the session for now, since i might want to add multisession per project later?

        # This triggers only when there is no grem.yaml file
        # as self.session_name is kept at None
        if self.session_name is None:
            self.session_name = session["name"]

        grem = Grem(name=self.session_name)

        for cur_window in session["windows"]:
            layout_value = cur_window.get("layout", None)
            window_obj = Window(cur_window["name"], Layout(layout_value))

            for cur_pane in cur_window["panes"]:
                pane_obj = Pane(cur_pane["command"], cwd=cur_pane["cwd"])
                window_obj.add_pane(pane_obj)

            grem.add_window(window_obj)

        return grem



from gremux.struct.context import Window, Pane, Layout
from gremux.struct.parser import Parser
from gremux.struct.logger import get_logger

__all__ = ["Window", "Pane", "Layout", "Parser", "get_logger"]



# ==============================================================================
# This file makes a rough implementation of tmux-sessionizer
# ==============================================================================

from pathlib import Path
from itertools import chain
from pyfzf.pyfzf import FzfPrompt
import libtmux

from gremux.core import get_places
import gremux.struct as gst


def sessionizer(logger):
    """
    tmux sessionizer
    Inspired by ThePrimeagen's tool
    https://github.com/ThePrimeagen/tmux-sessionizer

    When invoked, opens a fzf with all your common locations set up at
    ~/.config/gremux/places.yml

    Then opens a tmux session set up with the local grem.yaml file
    If there is not, then it opens a default setup
    """

    # get the common places and select the project directoy
    common_dirs = get_places()
    if common_dirs is None:
        message = [
            "places.yml is not configred! Run.",
            "gremux places create -s SOURCE",
            "Exiting",
        ]
        logger.info("\n".join(message))

        return None

    places = [Path(d).expanduser() for d in common_dirs]

    dirs = [
        p
        for p in chain.from_iterable(r.iterdir() for r in places if r.is_dir())
        if p.is_dir()
    ]

    selection = FzfPrompt().prompt(dirs)
    if not selection:
        logger.info("No directory selected, exiting...")
        return

    # connect to a tmux server

    server = libtmux.Server()

    proj_dir: str = selection[0]
    dir_name = Path(proj_dir).name

    # pass to the parser
    parser = gst.Parser(proj_dir)
    cfg: gst.Grem = parser.grem()

    # start the session with my 0th window
    match_session = server.sessions.filter(session_name=cfg.name)
    if len(match_session) > 0:
        match_session[0].attach(exit_=True)
        return None

    session = server.new_session(
        session_name=cfg.name,
        kill_session=True,  # replace if it already exists
        attach=False,
        window_name=cfg.windows[0].name,
        start_directory=proj_dir,
    )

    # setup and create rest of the windows
    for i, cfg_window in enumerate(cfg.windows):
        if i == 0:
            tmux_window = session.windows[0]
        else:
            tmux_window = session.new_window(window_name=cfg_window.name)

        cfg_window.apply(tmux_window, proj_dir)

    session.attach(exit_=True)

    return None



import logging
import sys


class ColorFormatter(logging.Formatter):
    COLORS = {
        logging.DEBUG: "\033[36m",  # cyan
        logging.INFO: "\033[32m",  # green
        logging.WARNING: "\033[33m",  # yellow
        logging.ERROR: "\033[31m",  # red
        logging.CRITICAL: "\033[41m",  # red background
    }
    RESET = "\033[0m"

    def format(self, record: logging.LogRecord) -> str:
        color = self.COLORS.get(record.levelno, "")
        levelname = record.levelname
        record.levelname = f"{color}{levelname}{self.RESET}"
        msg = super().format(record)
        record.levelname = levelname  # restore
        return msg


def get_logger(
    name: str = "gremux",
    level: int = logging.INFO,
) -> logging.Logger:
    """
    Get a logger object from the standard logging library that
    formats the text according to ColorFormatter

    Parameters:
    ----------
    * `name`: str
    * `level`: int
    """

    logger = logging.getLogger(name)
    logger.setLevel(level)

    # prevent duplicate handlers if called multiple times
    if logger.handlers:
        return logger

    handler = logging.StreamHandler(sys.stdout)
    formatter = ColorFormatter(fmt="[%(levelname)s] %(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    logger.propagate = False
    return logger



from dataclasses import dataclass, field
from typing import List
from gremux.struct.context import Window
import libtmux


@dataclass
class Grem:
    """
    Base class for the configuration file.
    """

    name: str
    windows: List[Window] = field(default_factory=list)

    def add_window(self, window: Window):
        """Add a Window to the configuration file"""
        self.windows.append(window)

    def launch(self, server: libtmux.Server, proj_dir):
        """
        Launch a tmux session with the current configuration

        Parameters:
        * `server`: libtimux.Server
        * `proj_dir`: str
        """
        # start the session with my 0th window
        match_session = server.sessions.filter(session_name=self.name)
        if len(match_session) > 0:
            match_session[0].attach(exit_=True)
            return None

        session = server.new_session(
            session_name=self.name,
            kill_session=True,  # replace if it already exists
            attach=False,
            window_name=self.windows[0].name,
            start_directory=proj_dir,
        )

        # setup and create rest of the windows
        for i, cfg_window in enumerate(self.windows):
            if i == 0:
                tmux_window = session.windows[0]
            else:
                tmux_window = session.new_window(window_name=cfg_window.name)

            cfg_window.apply(tmux_window, proj_dir)

        session.attach(exit_=True)

        return



import os
import yaml
import subprocess
import math
from gremux.struct.context import PlacesSource


def create(args, logger) -> None:
    if args.source is not None:
        create_source(args, logger)
    elif args.add is not None:
        create_add(args, logger)
    else:
        logger.info("Provide any flag --source or --add")

    return None


def create_source(args, logger) -> None:
    # a bit of validation from enum
    source = PlacesSource(args.source)

    if args.maximum is not None:
        max_items = int(args.maximum)
    else:
        # nothing will ever be greater than this
        max_items = math.inf

    home_dir = os.environ.get("HOME")
    places_file = os.path.join(home_dir, ".config", "gremux", "places.yaml")

    # ===== ZOXIDE SUPPORT ===== #
    if source == PlacesSource.ZOXIDE:
        result = subprocess.run(
            ["zoxide", "query", "--list"],
            capture_output=True,
            text=True,
            check=True,
        )
        result = result.stdout.strip().splitlines()

        if len(result) > max_items:
            result = result[:max_items]

        for path in result:
            logger.info(f"Added {path}")

        places = {"places": result}

    # ===== BACKUP SUPPORT ===== #
    elif source == PlacesSource.BACKUP:
        backup_file = os.path.join(home_dir, ".config", "gremux", "places.bak")
        with open(backup_file) as fh:
            places = yaml.safe_load(fh)

        for path in places["places"]:
            logger.info(f"Added {path}")

    # ===== DEFAULT PLACES ===== #
    elif source == PlacesSource.DEFAULT:
        places = {"places": [home_dir]}

    with open(places_file, "w") as fh:
        yaml.safe_dump(places, fh)

    logger.info(f"Written {places_file}")

    return None


def create_add(args, logger) -> None:
    home_dir = os.environ.get("HOME")
    places_file = os.path.join(home_dir, ".config", "gremux", "places.yaml")

    if not os.path.exists(places_file):
        message = [
            "places.yml is not configred! Run.",
            "gremux places create -s SOURCE",
            "Exiting",
        ]
        logger.info("\n".join(message))
        return None

    with open(places_file) as fh:
        places = yaml.safe_load(fh)

    for path in args.add:
        places["places"].append(path)
        logger.info(f"Added {path} to places.yaml")

    with open(places_file, "w") as fh:
        yaml.safe_dump(places, fh)

    logger.info(f"Written {places_file}")

    return None



import logging
from pathlib import Path
import gremux.struct as gst
import libtmux
import os
import yaml


def up(args, logger) -> None:
    if args.source is None:
        args.source = Path.cwd()

    up_source(args, logger)
    return None


def up_source(args, logger) -> None:
    parser = gst.Parser(args.source)
    cfg: gst.Grem = parser.grem()

    # connect to a tmux server

    server = libtmux.Server()

    cfg.launch(server, args.source)

    return None


def show(args, logger) -> None:
    if args.source is None:
        args.source = Path.cwd() / Path("grem.yaml")

    show_source(args, logger)
    return


def show_source(args, logger) -> None:
    if not Path(args.source).exists():
        logger.info("grem.yaml was not found. Exiting.")
        return

    with open(args.source) as fh:
        for line in fh:
            print(line.rstrip())

    return None


def create(args, logger) -> None:
    if args.source is None:
        logger.info("Must provide the --source argument")
        return None

    create_source(args, logger)

    return None


def create_source(args, logger) -> None:
    if args.source == "default":
        home_dir = os.environ.get("HOME")
        default_file = os.path.join(home_dir, ".config", "gremux", "default.yaml")

        default = {
            "session": {
                "name": "default",
                "windows": [
                    {
                        # ide
                        "name": "0",
                        "layout": None,
                        "panes": [
                            {
                                "cwd": ".",
                                "command": [None],
                            }
                        ],
                    },
                ],
            }
        }

        with open(default_file, "w") as fh:
            yaml.safe_dump(default, fh)

        logger.info(f"Written to {default_file}")
    else:
        logger.info("Feature not available. Exiting.")

    return None



import os
import yaml
from typing import List


def get_places() -> List:
    """
    Load the places.yaml file and return the list of common places in the system.
    """

    home_dir = os.environ.get("HOME")
    places_file = os.path.join(home_dir, ".config", "gremux", "places.yaml")

    if not os.path.exists(places_file):
        return None

    with open(places_file) as fh:
        places = yaml.safe_load(fh)

    return places["places"]



from dataclasses import dataclass, field
from typing import List
from enum import Enum

import os
import libtmux


class PlacesSource(Enum):
    ZOXIDE = "zoxide"
    DEFAULT = "default"
    BACKUP = "backup"
    NONE = None


class Layout(Enum):
    EVEN_HORIZONTAL = "even-horizontal"
    EVEN_VERTICAL = "even-vertical"
    MAIN_VERTICAL = "main-vertical"
    MAIN_HORIZONTAL = "main-horizontal"
    NONE = None


@dataclass
class Pane:
    """
    Base class for Pane configuration
    """

    command: str
    cwd: str | None = None
    focus: bool = False
    env: dict[str, str] = field(default_factory=dict)


@dataclass
class Window:
    """
    Base class for Window Configuration
    """

    name: str
    layout: Layout
    panes: List[Pane] = field(default_factory=list)

    def add_pane(self, pane: Pane):
        """
        Add a pane to the Window
        """
        self.panes.append(pane)

    def apply(self, tmux_window: libtmux.Window, proj_dir: str):
        """
        Apply the current Window configuration to a tmux Window

        Parameters:
        ----------
        * `tmux_window`: libtimux.Window
        * `proj_dir`: str
        """

        # ===== LAYOUT ===== #
        layout = self.layout.value
        if layout is not None:
            tmux_window.select_layout(layout)

        # ===== SPLITS ===== #
        layout = self.layout.value
        num_panes = len(self.panes)
        for _ in range(num_panes - 1):
            tmux_window.split()

        # ===== PANE COMMANDS ===== #

        # - move to correct dir - #
        for prior_pane, tg_pane in zip(tmux_window.panes, self.panes):
            # set the working dir
            if tg_pane.cwd is None:
                prior_pane.send_keys(f"cd {proj_dir}", enter=True)
            else:
                tg_dir = os.path.join(proj_dir, tg_pane.cwd)
                prior_pane.send_keys(f"cd {tg_dir}", enter=True)

            # prior_pane.send_keys("clear", enter=True)

            # - send respective commands - #
            if tg_pane.command is not None:
                # since there may be empty lists
                for cmd in tg_pane.command:
                    if cmd is not None:
                        # in case there are nulls
                        prior_pane.send_keys(cmd, enter=True)

            prior_pane.send_keys("clear", enter=True)

            # modify window width and all that

        # ===== MODIFY THE RELATIVE WIDTH/HEIGHT ===== #
        # wip
